#using SpinControl

"""
Given a magnetic strength h, find the gate fidelity for Pauli gates

e.g. X,Y,Z, sqrtX, sqrtY, sqrtZ

"""
function simulatedrivingfidelity(sim::Simulation)
    @assert sim.system isa SpinEnsemble
    @assert sim.dt>0
    @assert sim.h isa Real
    @assert sim.t2-sim.t1>π/sim.h #time resolution
    @assert sim.M>=sim.h^2 # NMC resolution 
    
    h=sim.h
    cluster = SpinCluster(sim.system)

    #swap t1 t2 to make sure of order
    if sim.t2<sim.t1
        sim.t1, sim.t2 = sim.t2, sim.t1
    end

    # find rabi frequency/time
    t=sim.t1:sim.dt:sim.t2 # for normal rabi with given resolution 
    if sim.M>=1e4
        tmp_M=round(Int, sqrt(sim.M)*10)
    else
        tmp_M=sim.M
    end
    curve=rabi(t, sim.system, h; M = tmp_M, N = 500, geterr = false)
    i_pi=argmin(curve) # time precision depends on dt!!


    label=[:X, :Y, :Z, :sqrtX, :sqrtY, :sqrtZ]
    data=zeros(sim.M,6) #initialize dataframe

    for i in 1:2 
        phase=(pi, pi/2)[i]
        Δt=phase/π * t[i_pi] # find evolution time different for h
        
        for j in 1:3
            aim=[0,0,0];  aim[j]=-1; # for half pi pulse 
            
            @showprogress for k in 1:sim.M
                ϕ_k, n_k = rabisampling(h, Δt, cluster, aim, N=sim.N)
                
                if phase==pi 
                    f=paulifidelity(ϕ_k, n_k, axis=j, phase=:_180)
                elseif phase==pi/2
                    f=paulifidelity(ϕ_k, n_k, axis=j, phase=:_90)
                else
                    error("invalid phase")
                end

                data[k, (i-1)*3+j]=f
                # reroll
                SpinControl.SpinEnsembles.reroll!(cluster)
            end
        end
    end

    return DataFrame(data, label)
end


