"""
Create time steps at logscale
"""
function logscaletime(t_max::Real, n_p::Int, dt::Real)
    @assert t_max > 1 && t_max > dt
    n = trunc(Int, 1 / dt)
    t1 = collect(dt:dt:n*dt)
    t2 = LinRange(0, log10(t_max), n_p - n)
    t2 = collect(map(x -> trunc(Int, x), (10 .^ t2) ./ dt)) .* dt
    return append!(t1, t2)
end

function logscaletime(t_min::Real, t_max::Real, n_p::Int, dt::Real)
    if t_min == 0
        return logscaletime(t_max, n_p, dt)
    end
    @assert t_min > 1 && t_min > dt
    t = LinRange(log10(t_min), log10(t_max), n_p)
    t = collect(map(x -> trunc(Int, x), (10 .^ t) ./ dt)) .* dt
    return t
end

## data generation
"""
Execution function that simulate logscale rabi envelope for given parameter pack.
"""
function simulatelogscalerabi(sim::Simulation)
    @assert sim.nt>0
    @assert sim.h isa Real
    
    dt = 2π/sim.h 
    t = logscaletime(sim.t1, sim.t2, sim.nt, dt)
    data = rabi(t, sim.system, sim.h; M = sim.M, N = sim.N, geterr = sim.geterr)
    if sim.geterr
        curve, err = data
        df = DataFrame(time = t, num_rabi = curve, num_err = err)
    else
        df = DataFrame(time = t, num_rabi = data)
    end
    return df
end


function simulatelogscalerabi(sim::Simulation, N::Int, λ::Real)
    @assert sim.nt>0
    @assert sim.h isa Real
    
    dt = 2π/sim.h 
    pre_t = logscaletime(sim.t1, sim.t2, sim.nt, dt)

    sim.nt = length(pre_t)
    t=vcat([LinRange(t_i-dt*λ,t_i,N+1) for t_i in pre_t]...)

    data = rabi(t, sim.system, sim.h; M = sim.M, N = sim.N, geterr = sim.geterr)

    if sim.geterr
        curve, err = data
        if getmax
            i_max=[(i-1)*(N+1)+argmax(curve[(i-1)*(N+1)+1:i*(N+1)]) for i in 1:sim.nt]
            t=t[i_max]
            curve=curve[i_max]
            err=err[i_max]
        end
        df = DataFrame(time = t, num_rabi = curve, num_err = err)
    else
        curve=data
        if getmax
            i_max=[(i-1)*(N+1)+argmax(curve[(i-1)*(N+1)+1:i*(N+1)]) for i in 1:sim.nt]
            t=t[i_max]
            curve=curve[i_max]
        end
        df = DataFrame(time = t, num_rabi = curve)
    end

    return df
end

"""
Execution function that simulate linear rabi envelope for given parameter pack.
"""
function simulaterabienvelope(sim::Simulation, N::Int, λ::Real, getmax::Bool=true)
    @assert sim.dt>0
    @assert sim.h isa Real
    
    dt = 2π/sim.h 
    pre_t = sim.t1:sim.dt:sim.t2

    sim.nt = length(pre_t)
    t=vcat([LinRange(t_i-dt*λ,t_i,N+1) for t_i in pre_t]...)

    data = rabi(t, sim.system, sim.h; M = sim.M, N = sim.N, geterr = sim.geterr)

    if sim.geterr
        curve, err = data
        if getmax
            i_max=[(i-1)*(N+1)+argmax(curve[(i-1)*(N+1)+1:i*(N+1)]) for i in 1:sim.nt]
            t=t[i_max]
            curve=curve[i_max]
            err=err[i_max]
        end
        df = DataFrame(time = t, num_rabi = curve, num_err = err)
    else
        curve=data
        if getmax
            i_max=[(i-1)*(N+1)+argmax(curve[(i-1)*(N+1)+1:i*(N+1)]) for i in 1:sim.nt]
            t=t[i_max]
            curve=curve[i_max]
        end
        df = DataFrame(time = t, num_rabi = curve)
    end

    return df
end
