using SpinControl

import CSV: CSV
import DataFrames: DataFrame
import Parameters: @with_kw
import Dates: now
import ProgressMeter: @showprogress
import JLD: load, save

const DATADIR = "./data/"
const OUTDIR = "./out/"
const METAFMT = ".siminfo"

# const GLOBAL_DATADIR="/home/zhang/OneDrive/Master_Thesis_Project/Simulations/data/"

"""
A set of parameters that are needed for all simulations. 
"""
@with_kw mutable struct Simulation
    system::Union{SpinEnsemble, SpinCluster} = ensemble3d
    protocol::Sequence = SquareSequence(0,0,[0])
    cycle::Int = 1
    
    h::Union{Real,Vector,Pulse,Sequence} = 0
    ρ::Matrix{<:Number} = [1.0+0im 0; 0 1.0+0im]
    
    # set time step
    t1::Real=0
    t2::Real=0
    dt::Real=0
    nt::Int=0
    # set monte carlo param
    M::Int=1
    N::Int=1
    geterr::Bool=true

    # output::Bool
    outdir::String = DATADIR
    # 
    cache::Bool = false

    description::String
    execution::Function ## output data
end

id(simulation::Simulation)=string(objectid(simulation), base=16)


function record(simulation::Simulation)
    metainfo = "
      Simulation Record  
      ===  
      id: $(id(simulation))  
      description: $(simulation.description) 
      date: $(now()) 
      system: $(simulation.system) 
      protocol: $(simulation.protocol)
      cycle: $(simulation.cycle)
      initial state: $(simulation.ρ)   
      magnetic field: $(simulation.h)  
      time range: $([simulation.t1, simulation.t2])   
      points: $(simulation.nt) 
      step: $(simulation.dt)  
      ensemble size: $(simulation.M)   
      sampling size: $(simulation.N)  
      error output: $(simulation.geterr)  
      data path: $(simulation.outdir) 
      cache: $(simulation.cache)
      execution function: $(simulation.execution)  
      ==="

    open(simulation.outdir * "simulation_" * id(simulation) * METAFMT, "w+") do file
        write(file, metainfo)
    end
end

"""
Read simualtion information of given id (hexadecimal string), print meta information
"""
function readsimulation(id::Union{AbstractString,UInt64}, sub_dir::AbstractString="";
    fmt::String = ".csv", verbose::Bool=true)
    if id isa UInt64
        id=string(id,base=16)
    end

    fn=DATADIR*sub_dir*"simulation_"*id
    if verbose
        open(fn*METAFMT, "r") do file 
            println(read(file, String))
        end
    end

    if fmt==".csv"
        return DataFrame(CSV.File(fn*fmt))
    elseif fmt==".jld"
        return load(fn*fmt, "data")
    end
end

function findidpattern(patterns::Vector{<:AbstractString}, data_dir::AbstractString="")
    for fn in readdir(data_dir, join=true)
        _, f_ext = splitext(fn)
        if f_ext==".siminfo"
            lines=readlines(fn)
            info=lines[5]
            if all([occursin(pattern, info) for pattern in patterns])
                return strip(lines[4])[5:end]
            end
        end
    end
end

"""
Save simulation data at given path and format
"""
function savedata(sim::Simulation, data::DataFrame)
    fn = sim.outdir * "simulation_$(id(sim))" * ".csv"
    CSV.write(fn, data)
end

function savedata(sim::Simulation, data::AbstractArray)
    fn = sim.outdir * "simulation_$(id(sim))" * ".jld"
    save(fn, "data", data)
end

function savedata(sim::Simulation, data::Tuple)
    fn = sim.outdir * "simulation_$(id(sim))" * ".jld"
    save(fn, "data", data)
end


"""
A pipeline function to execute the simulation and save its information in given directory. 
"""
function execute(simulation::Simulation)
    sim_id = id(simulation)
    println("Executing simulation: $sim_id")
    println(simulation.description)
    data=simulation.execution(simulation)
    
    println("Saving meta information...")
    record(simulation)
    println("Saving simulation data...")
    savedata(simulation, data)
    println("Done")
end

ensemble1d=SpinEnsemble(0.537188,1,[0,0,1],0.01,200,:spherical)
# ensemble1d=SpinEnsemble(0.537188,1,[0,0,1],0.05,200,:spherical)
ensemble2d=SpinEnsemble(0.377228,2,[0,0,1],0.05,30,:spherical)
ensemble3d=SpinEnsemble(0.39486,3,[0,0,1],0.1,10,:spherical)



# module
include("basics.jl")
include("rabienvelope.jl");
include("drivingerror.jl")
include("drivingfidelity.jl")
include("sequence.jl")
